# -*- coding: utf-8 -*-
#
# Copyright (c) 2015, Alcatel-Lucent Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its contributors
#       may be used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import sys
from builtins import next
from builtins import object
import weakref
import datetime
from uuid import uuid4
from copy import deepcopy

from bambou import bambou_logger
from .exceptions import BambouHTTPError, InternalConsitencyError
from .nurest_connection import NURESTConnection, HTTP_METHOD_DELETE, HTTP_METHOD_PUT, HTTP_METHOD_POST, HTTP_METHOD_GET
from .nurest_request import NURESTRequest
from .nurest_session import NURESTSession
from .utils import NURemoteAttribute
from .utils.decorators import backwards_compatible_async
from .config import BambouConfig
from future.utils import with_metaclass


class NUMetaRESTObject(type):  # pragma: no cover
    """
    """
    @property
    def rest_name(cls):
        """ Represents a singular REST name
        """
        if cls.__name__ == "NURESTRootObject" or cls.__name__ == "NURESTObject":
            return "Not Implemented"

        if cls.__rest_name__ is None:
            raise NotImplementedError('%s has no defined name. Implement rest_name property first.' % cls)

        return cls.__rest_name__

    @property
    def resource_name(cls):
        """ Represents the resource name
        """
        if cls.__name__ == "NURESTRootObject" or cls.__name__ == "NURESTObject":
            return "Not Implemented"

        if cls.__resource_name__ is None:
            raise NotImplementedError('%s has no defined resource name. Implement resource_name property first.' % cls)

        return cls.__resource_name__


class NURESTObject(with_metaclass(NUMetaRESTObject, object)):
    """ Determines an object as a NURESTObject one
        Provides basic saving and fetching utilities
    """
    __rest_name__ = None
    __resource_name__ = None

    def __init__(self):
        """ Initializes the object with general information

            Args:
                creation_date: float representing time since epoch
                id: identifier of the object
                local_id: internal identifier of the object
                owner: string representing the owner
                parent_id: identifier of the object's parent
                parent_type: type of the parent
        """

        self._local_id = str(uuid4())
        self._creation_date = None
        self._last_updated_date = None
        self._id = None
        self._owner = None
        self._parent_id = None
        self._parent_type = None
        self._parent = None
        self._is_dirty = False

        self._attribute_errors = dict()
        self._attributes = dict()

        self.expose_attribute(local_name='id', remote_name=BambouConfig.get_id_remote_name(), attribute_type=BambouConfig.get_id_type(), is_identifier=True)
        self.expose_attribute(local_name='parent_id', remote_name='parentID', attribute_type=str)
        self.expose_attribute(local_name='parent_type', remote_name='parentType', attribute_type=str)
        self.expose_attribute(local_name='creation_date', remote_name='creationDate', attribute_type=float, is_editable=False)
        self.expose_attribute(local_name='last_updated_date', remote_name='lastUpdatedDate', attribute_type=float, is_editable=False)
        self.expose_attribute(local_name='owner', attribute_type=str, is_readonly=True)

        self._fetchers_registry = dict()

    def _compute_args(self, data=dict(), **kwargs):
        """ Compute the arguments

            Try to import attributes from data.
            Otherwise compute kwargs arguments.

            Args:
                data: a dict()
                kwargs: a list of arguments
        """

        for name, remote_attribute in self._attributes.items():
            default_value = BambouConfig.get_default_attribute_value(self.__class__, name, remote_attribute.attribute_type)
            setattr(self, name, default_value)

        if len(data) > 0:
            self.from_dict(data)

        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)

    # Properties

    @property
    def creation_date(self):
        """ Get creation date """

        return self._creation_date

    @creation_date.setter
    def creation_date(self, creation_date):
        """ Set creation date """

        self._creation_date = creation_date

    @property
    def last_updated_date(self):
        """ Get last update date """

        return self._last_updated_date

    @last_updated_date.setter
    def last_updated_date(self, update_date):
        """ Set creation date """

        self._last_updated_date = update_date

    @property
    def id(self):
        """ Get object id """

        return self._id

    @id.setter
    def id(self, id):
        """ Set object id """

        self._id = id

    @property
    def local_id(self):
        """ Get local id """

        return self._local_id

    @property
    def owner(self):
        """ Get owner """

        return self._owner

    @owner.setter
    def owner(self, owner):
        """ Set owner """

        self._owner = owner

    @property
    def parent_id(self):
        """ Get parent id """

        return self._parent_id

    @parent_id.setter
    def parent_id(self, parent_id):
        """ Set parent id """

        self._parent_id = parent_id

    @property
    def parent_type(self):
        """ Get parent type """

        return self._parent_type

    @parent_type.setter
    def parent_type(self, parent_type):
        """ Set parent type """

        self._parent_type = parent_type

    @property
    def parent_object(self):
        """ Get parent """

        return self._parent() if self._parent else None

    @parent_object.setter
    def parent_object(self, parent):
        """ Set parent id """

        self._parent = weakref.ref(parent) if parent else None

    @property
    def rest_name(self):
        """ Returns the current ReST name of the object.

            Returns:
                Returns a dictionnary containing attribute information
        """
        return self.__class__.rest_name

    @property
    def resource_name(self):
        """ Resource name of the object.

            It will compute the plural if needed

            Returns:
                Returns a string that represents the resouce name of the object
        """
        return self.__class__.resource_name

    @property
    def fetchers(self):
        """ Return a copy of all fetchers

            Returns:
                list: list of all fetchers

            Example:
                >>> print entity.fetchers
                [<NUSubEntitiesFetcher at xxxx>, <NUOtherEntitiesFetcher at yyyy>]
        """
        return deepcopy(list(self._fetchers_registry.values()))

    # Children

    @property
    def children_rest_names(self):
        """ Gets the list of all possible children ReST names.

            Returns:
                list: list containing all possible rest names as string

            Example:
                >>> entity = NUEntity()
                >>> entity.children_rest_names
                ["foo", "bar"]
        """

        names = []

        for fetcher in self.fetchers:
            names.append(fetcher.__class__.managed_object_rest_name())

        return names

    # Class methods

    @classmethod
    def rest_base_url(cls):
        """ Override this method to set object base url """

        controller = NURESTSession.get_current_session().login_controller
        return controller.url

    @classmethod
    def object_with_id(cls, id):
        """ Get a new object with the given id

            Returns:
                Returns a new instance with specified id
        """

        new_object = cls()
        new_object.id = id

        return new_object

    # URL and resource management

    def get_resource_url(self):
        """ Get resource complete url """

        name = self.__class__.resource_name
        url = self.__class__.rest_base_url()

        if self.id is not None:
            return "%s/%s/%s" % (url, name, self.id)

        return "%s/%s" % (url, name)

    def get_resource_url_for_child_type(self, nurest_object_type):
        """ Get the resource url for the nurest_object type """

        return "%s/%s" % (self.get_resource_url(), nurest_object_type.resource_name)

    def __str__(self):
        """ Prints a NURESTObject """

        return "%s (ID=%s)" % (self.__class__, self.id)

    def validate(self):
        """ Validate the current object attributes.

            Check all attributes and store errors

            Returns:
                Returns True if all attibutes of the object
                respect contraints. Returns False otherwise and
                store error in errors dict.

        """
        self._attribute_errors = dict()  # Reset validation errors

        for local_name, attribute in self._attributes.items():

            value = getattr(self, local_name, None)

            if attribute.is_required and (value is None or value == ""):
                self._attribute_errors[local_name] = {'title': 'Invalid input',
                                                      'description': 'This value is mandatory.',
                                                      'remote_name': attribute.remote_name}
                continue

            if value is None:
                continue  # without error

            if not self._validate_type(local_name, attribute.remote_name, value, attribute.attribute_type):
                continue

            if attribute.min_length is not None and len(value) < attribute.min_length:
                self._attribute_errors[local_name] = {'title': 'Invalid length',
                                                      'description': 'Attribute %s minimum length should be %s but is %s' % (attribute.remote_name, attribute.min_length, len(value)),
                                                      'remote_name': attribute.remote_name}
                continue

            if attribute.max_length is not None and len(value) > attribute.max_length:
                self._attribute_errors[local_name] = {'title': 'Invalid length',
                                                      'description': 'Attribute %s maximum length should be %s but is %s' % (attribute.remote_name, attribute.max_length, len(value)),
                                                      'remote_name': attribute.remote_name}
                continue

            if attribute.attribute_type == list:
                valid = True
                for item in value:
                    if valid is True:
                        valid = self._validate_value(local_name, attribute, item)
            else:
                self._validate_value(local_name, attribute, value)

        return self.is_valid()

    def _validate_type(self, local_name, remote_name, value, validate_type):
        if type(value) != validate_type:
            # On python 2, we accept unicode input when attribute_type is set to str
            if not (sys.version_info < (3,) and validate_type == str and type(value) == unicode):
                self._attribute_errors[local_name] = {'title': 'Wrong type',
                                                      'description': 'Attribute %s type should be %s but is %s' % (remote_name, validate_type, type(value)),
                                                      'remote_name': remote_name}
                return False

        return True

    def _validate_value(self, local_name, attribute, value):

        if attribute.subtype is not None and not self._validate_type(local_name, attribute.remote_name, value, attribute.subtype):
            return False

        if attribute.min_value is not None and value < attribute.min_value:
            self._attribute_errors[local_name] = {'title': 'Invalid value',
                                                  'description':
                                                      'Attribute %s minimum value should be %s but is %s' %
                                                      (attribute.remote_name, attribute.min_value, value),
                                                  'remote_name': attribute.remote_name}
            return False

        if attribute.max_value is not None and value > attribute.max_value:
            self._attribute_errors[local_name] = {'title': 'Invalid value',
                                                  'description':
                                                      'Attribute %s maximum value should be %s but is %s' %
                                                      (attribute.remote_name, attribute.max_value, value),
                                                  'remote_name': attribute.remote_name}
            return False

        if attribute.choices and value not in attribute.choices:
            self._attribute_errors[local_name] = {'title': 'Invalid input',
                                                  'description': 'Value %s not a valid choice' % value,
                                                  'remote_name': attribute.remote_name}
            return False

        return True

    def is_valid(self):
        """
        """
        return len(self.errors) == 0

    @property
    def errors(self):
        """
        """
        return self._attribute_errors

    def expose_attribute(self, local_name, attribute_type, remote_name=None, display_name=None, is_required=False, is_readonly=False, max_length=None, min_length=None, is_identifier=False, choices=None, is_unique=False, is_email=False, is_login=False, is_editable=True, is_password=False, can_order=False, can_search=False, subtype=None, min_value=None, max_value=None):
        """ Expose local_name as remote_name

            An exposed attribute `local_name` will be sent within the HTTP request as
            a `remote_name`

        """
        if remote_name is None:
            remote_name = local_name

        if display_name is None:
            display_name = local_name

        attribute = NURemoteAttribute(local_name=local_name, remote_name=remote_name, attribute_type=attribute_type)
        attribute.display_name = display_name
        attribute.is_required = is_required
        attribute.is_readonly = is_readonly
        attribute.min_length = min_length
        attribute.max_length = max_length
        attribute.is_editable = is_editable
        attribute.is_identifier = is_identifier
        attribute.choices = choices
        attribute.is_unique = is_unique
        attribute.is_email = is_email
        attribute.is_login = is_login
        attribute.is_password = is_password
        attribute.can_order = can_order
        attribute.can_search = can_search
        attribute.subtype = subtype
        attribute.min_value = min_value
        attribute.max_value = max_value

        self._attributes[local_name] = attribute

    def get_attributes(self):
        """ Get all attributes information

            Returns:
                Returns a dictionnary containing attribute information
        """

        return list(self._attributes.values())

    def get_attribute_infos(self, local_name):
        """ Get exposed attribute information

            Args:
                local_name: the attribute name

            Returns:
                A dictionary of all information

        """
        if local_name in self._attributes:
            return self._attributes[local_name]

        return None

    def is_owned_by_current_user(self):
        """ Check if the current user owns the object """

        from bambou.nurest_root_object import NURESTRootObject
        root_object = NURESTRootObject.get_default_root_object()
        return self._owner == root_object.id

    def parent_for_matching_rest_name(self, rest_names):
        """ Return parent that matches a rest name """

        parent = self

        while parent:
            if parent.rest_name in rest_names:
                return parent

            parent = parent.parent_object

        return None

    def genealogic_types(self):
        """ Get genealogic types

            Returns:
                Returns a list of all parent types
        """

        types = []
        parent = self

        while parent:
            types.append(parent.rest_name)
            parent = parent.parent_object

        return types

    def genealogic_ids(self):
        """ Get all genealogic ids

            Returns:
                 A list of all parent ids
        """

        ids = []
        parent = self

        while parent:
            ids.append(parent.id)
            parent = parent.parent_object

        return ids

    def genealogy_contains_type(self, resource_name):
        """ Check if parents contains an object of type resource_name

            Args:
                resource_name: the name of the resource to find

            Returns:
                Returns True if a parent of type has been found. False otherwise.
        """

        resource_names = self.genealogic_types()
        return resource_name in resource_names

    def genealogy_contains_id(self, id):
        """ Check if parents contains an object of type resource_name

            Args:
                id: the id of the resource to find

            Returns:
                Returns True if a parent with specific id has been found. False otherwise.
        """

        ids = self.genealogic_ids()
        return id in ids

    def get_formated_creation_date(self, format='%b %Y %d %H:%I:%S'):
        """ Return creation date with a given format. Default is '%b %Y %d %H:%I:%S' """

        if not self._creation_date:
            return None

        date = datetime.datetime.utcfromtimestamp(self._creation_date)
        return date.strftime(format)

    # Fetchers registry

    def register_fetcher(self, fetcher, rest_name):
        """ Register a children fetcher

        """
        self._fetchers_registry[rest_name] = fetcher

    def fetcher_for_rest_name(self, rest_name):
        """ Returns the children fetcher for the given rest name

            Args:
                rest_name (string): the children rest name

            Returns:
                list: Returns the corresponding fetcher

            Example:
                >>> print entity.fetcher_for_rest_name(NUSubEntity.rest_name)
                <NUSubEntitiesFetcher at yyyy>
        """
        if rest_name not in self._fetchers_registry:
            return None

        return self._fetchers_registry[rest_name]

    # Children management

    def add_child(self, child):
        """ Add a child """

        rest_name = child.rest_name
        children = self.fetcher_for_rest_name(rest_name)

        if children is None:
            raise InternalConsitencyError('Could not find fetcher with name %s while adding %s in parent %s' % (rest_name, child, self))

        if child not in children:
            child.parent_object = self
            children.append(child)

    def remove_child(self, child):
        """ Remove a child """

        rest_name = child.rest_name
        children = self.fetcher_for_rest_name(rest_name)

        target_child = None

        for local_child in children:
            if local_child.id == child.id:
                target_child = local_child
                break

        if target_child:
            target_child.parent_object = None
            children.remove(target_child)

    def update_child(self, child):
        """ Update child """

        rest_name = child.rest_name
        children = self.fetcher_for_rest_name(rest_name)

        index = None

        for local_child in children:
            if local_child.id == child.id:
                index = children.index(local_child)
                break

        if index is not None:
            children[index] = child

    # Compression / Decompression

    def copy(self):
        """ Returns a copy of the current object

            Returns:
                A copy of the given object

            Example::
                >>> print entity.to_dict()
                <Entity object at 0x10b9d4910>

        """

        return self.__class__(data=self.to_dict())

    def to_dict(self):
        """ Converts the current object into a Dictionary using all exposed ReST attributes.

            Returns:
                dict: the dictionary containing all the exposed ReST attributes and their values.

            Example::
                >>> print entity.to_dict()
                {"name": "my entity", "description": "Hello World", "ID": "xxxx-xxx-xxxx-xxx", ...}
        """

        dictionary = dict()

        for local_name, attribute in self._attributes.items():
            remote_name = attribute.remote_name

            if hasattr(self, local_name):
                value = getattr(self, local_name)

                # Removed to resolve issue http://mvjira.mv.usa.alcatel.com/browse/VSD-5940 (12/15/2014)
                # if isinstance(value, bool):
                #     value = int(value)

                if isinstance(value, NURESTObject):
                    value = value.to_dict()

                if isinstance(value, list) and len(value) > 0 and isinstance(value[0], NURESTObject):
                    tmp = list()
                    for obj in value:
                        tmp.append(obj.to_dict())

                    value = tmp

                dictionary[remote_name] = value
            else:
                pass  # pragma: no cover

        return dictionary

    def from_dict(self, dictionary):
        """ Sets all the exposed ReST attribues from the given dictionary

            Args:
                dictionary (dict): dictionnary containing the raw object attributes and their values.

            Example:
                >>> info = {"name": "my group", "private": False}
                >>> group = NUGroup()
                >>> group.from_dict(info)
                >>> print "name: %s - private: %s" % (group.name, group.private)
                "name: my group - private: False"
        """

        for remote_name, remote_value in dictionary.items():
            # Check if a local attribute is exposed with the remote_name
            # if no attribute is exposed, return None
            local_name = next((name for name, attribute in self._attributes.items() if attribute.remote_name == remote_name), None)

            if local_name:
                setattr(self, local_name, remote_value)
            else:
                # print('Attribute %s could not be added to object %s' % (remote_name, self))
                pass  # pragma: no cover

    # HTTP Calls

    @backwards_compatible_async
    def delete(self, response_choice=1, as_async=False, callback=None):
        """ Delete object and call given callback in case of call.

            Args:
                response_choice (int): Automatically send a response choice when confirmation is needed
                as_async (bool): Boolean to make an asynchronous call. Default is False
                callback (function): Callback method that will be triggered in case of asynchronous call

            Example:
                >>> entity.delete() # will delete the enterprise from the server
        """
        return self._manage_child_object(nurest_object=self, method=HTTP_METHOD_DELETE, as_async=as_async, callback=callback, response_choice=response_choice)

    @backwards_compatible_async
    def save(self, response_choice=None, as_async=False, callback=None):
        """ Update object and call given callback in case of as_async call

            Args:
                as_async (bool): Boolean to make an asynchronous call. Default is False
                callback (function): Callback method that will be triggered in case of asynchronous call

            Example:
                >>> entity.name = "My Super Object"
                >>> entity.save() # will save the new name in the server
        """
        return self._manage_child_object(nurest_object=self, method=HTTP_METHOD_PUT, as_async=as_async, callback=callback, response_choice=response_choice)

    @backwards_compatible_async
    def fetch(self, as_async=False, callback=None):
        """ Fetch all information about the current object

            Args:
                as_async (bool): Boolean to make an asynchronous call. Default is False
                callback (function): Callback method that will be triggered in case of asynchronous call

            Returns:
                tuple: (current_fetcher, callee_parent, fetched_bjects, connection)

            Example:
                >>> entity = NUEntity(id="xxx-xxx-xxx-xxx")
                >>> entity.fetch() # will get the entity with id "xxx-xxx-xxx-xxx"
                >>> print entity.name
                "My Entity"
        """

        if self.id is None:
            raise InternalConsitencyError("Cannot fetch an object that does not have an ID")

        request = NURESTRequest(method=HTTP_METHOD_GET, url=self.get_resource_url())

        if as_async:
            return self.send_request(request=request, as_async=as_async, local_callback=self._did_retrieve, remote_callback=callback)
        else:
            connection = self.send_request(request=request)
            return self._did_retrieve(connection)

    # REST HTTP Calls

    @backwards_compatible_async
    def send_request(self, request, as_async=False, local_callback=None, remote_callback=None, user_info=None):
        """ Sends a request, calls the local callback, then the remote callback in case of as_async call

            Args:
                request: The request to send
                local_callback: local method that will be triggered in case of as_async call
                remote_callback: remote method that will be triggered in case of as_async call
                user_info: contains additionnal information to carry during the request

            Returns:
                Returns the object and connection (object, connection)
        """

        callbacks = dict()

        if local_callback:
            callbacks['local'] = local_callback

        if remote_callback:
            callbacks['remote'] = remote_callback

        connection = NURESTConnection(request=request, as_async=as_async, callback=self._did_receive_response, callbacks=callbacks)
        connection.user_info = user_info

        return connection.start()

    def _manage_child_object(self, nurest_object, method=HTTP_METHOD_GET, as_async=False, callback=None, handler=None, response_choice=None, commit=False):
        """ Low level child management. Send given HTTP method with given nurest_object to given ressource of current object

            Args:
                nurest_object: the NURESTObject object to manage
                method: the HTTP method to use (GET, POST, PUT, DELETE)
                callback: the callback to call at the end
                handler: a custom handler to call when complete, before calling the callback
                commit: True to auto commit changes in the current object

            Returns:
                Returns the object and connection (object, connection)
        """
        url = None

        if method == HTTP_METHOD_POST:
            url = self.get_resource_url_for_child_type(nurest_object.__class__)
        else:
            url = self.get_resource_url()

        if response_choice is not None:
            url += '?responseChoice=%s' % response_choice

        request = NURESTRequest(method=method, url=url, data=nurest_object.to_dict())
        user_info = {'nurest_object': nurest_object, 'commit': commit}

        if not handler:
            handler = self._did_perform_standard_operation

        if as_async:
            return self.send_request(request=request, as_async=as_async, local_callback=handler, remote_callback=callback, user_info=user_info)
        else:
            connection = self.send_request(request=request, user_info=user_info)
            return handler(connection)

    def _manage_children_objects(self, nurest_objects, method=HTTP_METHOD_GET, as_async=False, callback=None, handler=None, response_choice=None, commit=False):
        """ Low level bulk child management. Send given HTTP method with given nurest_objects to given ressource of current object

            Args:
                nurest_objects: the list of NURESTObject objects to manage
                method: the HTTP method to use (GET, POST, PUT, DELETE)
                callback: the callback to call at the end
                handler: a custom handler to call when complete, before calling the callback
                commit: True to auto commit changes in the current object

            Returns:
                Returns the object and connection (object, connection)
        """
        url = None

        if type(nurest_objects) is not list or not nurest_objects:
            raise TypeError('This method expects a list of objects, with at least a single object provided.')

        if method in [HTTP_METHOD_POST, HTTP_METHOD_PUT, HTTP_METHOD_DELETE]:
            # HTTP_METHOD_PUT & HTTP_METHOD_DELETE should only come in from nurest_root_object,
            # in which case the get_resource_url_for_child_type will provide the root path URL, not the parent based one.
            url = self.get_resource_url_for_child_type(nurest_objects[0].__class__)
        else:
            # Backup in case other methods are ever used for bulk operations, unlikely
            url = self.get_resource_url()

        if response_choice is not None:
            url += '?responseChoice=%s' % response_choice
        
        if method == HTTP_METHOD_DELETE:
            for nurest_object in nurest_objects:
                if nurest_object.id:
                    if '?' not in url:
                        url += '?id={0:s}'.format(nurest_object.id)
                    else:
                        url += '&id={0:s}'.format(nurest_object.id)

        data_object = list()
        for nurest_object in nurest_objects:
            data_object.append(nurest_object.to_dict())
        request = NURESTRequest(method=method, url=url, data=data_object)
        user_info = {'nurest_objects': nurest_objects, 'commit': commit}

        if not handler:
            handler = self._did_perform_bulk_operation

        if as_async:
            return self.send_request(request=request, as_async=as_async, local_callback=handler, remote_callback=callback, user_info=user_info)
        else:
            connection = self.send_request(request=request, user_info=user_info)
            return handler(connection)

    # REST Operation handlers

    def _did_receive_response(self, connection):
        """ Receive a response from the connection """

        if connection.has_timeouted:
            bambou_logger.info("NURESTConnection has timeout.")
            return

        has_callbacks = connection.has_callbacks()
        should_post = not has_callbacks

        if connection.handle_response_for_connection(should_post=should_post) and has_callbacks:
            callback = connection.callbacks['local']
            callback(connection)

    def _did_retrieve(self, connection):
        """ Callback called after fetching the object """

        response = connection.response

        try:
            self.from_dict(response.data[0])
        except:
            pass

        return self._did_perform_standard_operation(connection)

    def _did_perform_standard_operation(self, connection):
        """ Performs standard opertions """

        if connection.as_async:
            callback = connection.callbacks['remote']

            if connection.user_info and 'nurest_object' in connection.user_info:
                callback(connection.user_info['nurest_object'], connection)
            else:
                callback(self, connection)
        else:
            if connection.response.status_code >= 400 and BambouConfig._should_raise_bambou_http_error:
                raise BambouHTTPError(connection=connection)

            # Case with multiple objects like assignment
            if connection.user_info and 'nurest_objects' in connection.user_info:

                if connection.user_info['commit']:
                    for nurest_object in connection.user_info['nurest_objects']:
                        self.add_child(nurest_object)
                return (connection.user_info['nurest_objects'], connection)

            if connection.user_info and 'nurest_object' in connection.user_info:

                if connection.user_info['commit']:
                    self.add_child(connection.user_info['nurest_object'])

                return (connection.user_info['nurest_object'], connection)

            return (self, connection)

    def _did_perform_bulk_operation(self, connection, errors=None):
        """ Performs bulk opertions """

        if not errors:
            response = connection.response
            errors = list()
            try:
                obj_index = 0
                for nurest_object in connection.user_info['nurest_objects']:
                    if len(response.data) > obj_index and 'status' in response.data[obj_index] and response.data[obj_index]['status'] >= 400:
                        errors.append({
                            "object": nurest_object,
                            "index": obj_index,
                            "status": response.data[obj_index]['status'],
                            "error": response.data[obj_index]['data'] if 'data' in response.data[obj_index] else {}
                        })
                    elif len(response.data) <= obj_index:
                        errors.append({
                            "object": nurest_object,
                            "index": obj_index,
                            "status": None,
                            "error": {
                                "errors": "No return data presented for object, unhandled"
                            }
                        })
                    obj_index += 1
            except Exception:
                pass
        
            if not errors:
                errors = None

        if connection.as_async:
            callback = connection.callbacks['remote']

            if connection.user_info and 'nurest_objects' in connection.user_info:
                callback(connection.user_info['nurest_objects'], connection, errors)
            else:
                callback(self, connection, errors)
        else:
            if connection.response.status_code >= 400 and BambouConfig._should_raise_bambou_http_error:
                raise BambouHTTPError(connection=connection)

            if connection.user_info and 'nurest_objects' in connection.user_info:

                if connection.user_info['commit']:
                    for nurest_object in connection.user_info['nurest_objects']:
                        if nurest_object.id:
                            self.add_child(nurest_object)
                return (connection.user_info['nurest_objects'], connection, errors)

            return (self, connection, errors)

    # Advanced REST Operations

    @backwards_compatible_async
    def create_child(self, nurest_object, response_choice=None, as_async=False, callback=None, commit=True):
        """ Add given nurest_object to the current object

            For example, to add a child into a parent, you can call
            parent.create_child(nurest_object=child)

            Args:
                nurest_object (bambou.NURESTObject): the NURESTObject object to add
                response_choice (int): Automatically send a response choice when confirmation is needed
                as_async (bool): should the request be done asynchronously or not
                callback (function): callback containing the object and the connection

            Returns:
                Returns the object and connection (object, connection)

            Example:
                >>> entity = NUEntity(name="Super Entity")
                >>> parent_entity.create_child(entity) # the new entity as been created in the parent_entity
        """

        # if nurest_object.id:
        #     raise InternalConsitencyError("Cannot create a child that already has an ID: %s." % nurest_object)

        return self._manage_child_object(nurest_object=nurest_object,
                                         as_async=as_async,
                                         method=HTTP_METHOD_POST,
                                         callback=callback,
                                         handler=self._did_create_child,
                                         response_choice=response_choice,
                                         commit=commit)

    @backwards_compatible_async
    def create_children(self, nurest_objects, response_choice=None, as_async=False, callback=None, commit=True):
        """ Add a list of nurest_objects to the current object

            For example, to add multiple children of the same type into a parent, you can call
            parent.create_children(nurest_objects=[child_a, child_b, child_c])

            Args:
                nurest_objects ([bambou2.NURESTObject]): the list of NURESTObject objects to add
                response_choice (int): Automatically send a response choice when confirmation is needed
                as_async (bool): should the request be done asynchronously or not
                callback (function): callback containing the object and the connection

            Returns:
                Returns the object and connection (object, connection)

            Example:
                >>> entity = NUEntity(name="Super Entity")
                >>> parent_entity.create_child(entity) # the new entity as been created in the parent_entity
        """

        if not nurest_objects:
            raise InternalConsitencyError("One or more objects have to be provided.")

        return self._manage_children_objects(nurest_objects=nurest_objects,
                                             as_async=as_async,
                                             method=HTTP_METHOD_POST,
                                             callback=callback,
                                             handler=self._did_create_children,
                                             response_choice=response_choice,
                                             commit=commit)

    @backwards_compatible_async
    def instantiate_child(self, nurest_object, from_template, response_choice=None, as_async=False, callback=None, commit=True):
        """ Instantiate an nurest_object from a template object

            Args:
                nurest_object: the NURESTObject object to add
                from_template: the NURESTObject template object
                callback: callback containing the object and the connection

            Returns:
                Returns the object and connection (object, connection)

            Example:
                >>> parent_entity = NUParentEntity(id="xxxx-xxxx-xxx-xxxx") # create a NUParentEntity with an existing ID (or retrieve one)
                >>> other_entity_template = NUOtherEntityTemplate(id="yyyy-yyyy-yyyy-yyyy") # create a NUOtherEntityTemplate with an existing ID (or retrieve one)
                >>> other_entity_instance = NUOtherEntityInstance(name="my new instance") # create a new NUOtherEntityInstance to be intantiated from other_entity_template
                >>>
                >>> parent_entity.instantiate_child(other_entity_instance, other_entity_template) # instatiate the new domain in the server
        """

        # if nurest_object.id:
        #     raise InternalConsitencyError("Cannot instantiate a child that already has an ID: %s." % nurest_object)

        if not from_template.id:
            raise InternalConsitencyError("Cannot instantiate a child from a template with no ID: %s." % from_template)

        nurest_object.template_id = from_template.id
        return self._manage_child_object(nurest_object=nurest_object,
                                         as_async=as_async,
                                         method=HTTP_METHOD_POST,
                                         callback=callback,
                                         handler=self._did_create_child,
                                         response_choice=response_choice,
                                         commit=commit)

    def _did_create_child(self, connection):
        """ Callback called after adding a new child nurest_object """

        response = connection.response
        try:
            connection.user_info['nurest_object'].from_dict(response.data[0])
        except Exception:
            pass

        return self._did_perform_standard_operation(connection)

    def _did_create_children(self, connection):
        """ Callback called after adding a new children nurest_objects """

        response = connection.response
        errors = list()
        try:
            obj_index = 0
            for nurest_object in connection.user_info['nurest_objects']:
                if len(response.data) > obj_index and 'status' in response.data[obj_index] and response.data[obj_index]['status'] < 300 and 'data' in response.data[obj_index]:
                    nurest_object.from_dict(response.data[obj_index]['data'])
                obj_index += 1
        except Exception:
            pass
        
        if not errors:
            errors = None

        return self._did_perform_bulk_operation(connection, errors)

    @backwards_compatible_async
    def assign(self, objects, nurest_object_type, as_async=False, callback=None, commit=True):
        """ Reference a list of objects into the current resource

            Args:
                objects (list): list of NURESTObject to link
                nurest_object_type (type): Type of the object to link
                callback (function): Callback method that should be fired at the end

            Returns:
                Returns the current object and the connection (object, connection)

            Example:
                >>> entity.assign([entity1, entity2, entity3], NUEntity) # entity1, entity2 and entity3 are now part of the entity
        """

        ids = list()

        for nurest_object in objects:
            ids.append(nurest_object.id)

        url = self.get_resource_url_for_child_type(nurest_object_type)
        request = NURESTRequest(method=HTTP_METHOD_PUT, url=url, data=ids)
        user_info = {'nurest_objects': objects, 'commit': commit}

        if as_async:
            return self.send_request(request=request,
                                     as_async=as_async,
                                     local_callback=self._did_perform_standard_operation,
                                     remote_callback=callback,
                                     user_info=user_info)
        else:
            connection = self.send_request(request=request,
                                           user_info=user_info)

            return self._did_perform_standard_operation(connection)

    # Comparison

    def rest_equals(self, rest_object):
        """ Compare objects REST attributes

        """
        if not self.equals(rest_object):
            return False

        return self.to_dict() == rest_object.to_dict()

    def equals(self, rest_object):
        """ Compare with another object """

        if self._is_dirty:
            return False

        if rest_object is None:
            return False

        if not isinstance(rest_object, NURESTObject):
            raise TypeError('The object is not a NURESTObject %s' % rest_object)

        if self.rest_name != rest_object.rest_name:
            return False

        if self.id and rest_object.id:
            return self.id == rest_object.id

        if self.local_id and rest_object.local_id:
            return self.local_id == rest_object.local_id

        return False



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015, Alcatel-Lucent Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its contributors
#       may be used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import print_function

__doc__ = """
# General Concepts


## Introduction

Bambou provides a set of objects that allow the manipulation of ReST entities very easily. It deals with all possible CRUD operations.
It is based on the library `Bambou`, which defines all these low level operations in a single place.

`Bambou` is composed of the following important classes:

* `bambou.NURESTSession`
    Class representing an authenticated session.

* `bambou.NURESTObject`
    Parent class of all ReST entities. All ReST exposed object objects inherit from this class.

* `bambou.NURESTFetcher`
    Class used to get children of a `bambou.NURESTObject`.

* `bambou.NURESTPushCenter`
    Class that deals with intercepting and rerouting ReST Push Notifications.

> There are more objects in `Bambou`, but you don't need to know all of them for now.



## NURESTSession

The `bambou.NURESTSession` represents some user credentials coupled with an API URL. All ReST calls are done using
the current active session. `bambou.NURESTSession` is an abstract class that must be reimplemented by anything using `Bambou`.

In a `MySDK` using bambou, you use a class named `mysdk.v3_2.MySession` which will be used in the following examples.


    #!python
    session = My}Session(username="user", password="secret", enterprise="organization", api_url="https://server")
    session.start()

    # your script

When you start the session, a ReST call will be sent to the API endpoint in order to get the API key.
If the credentials are valid, the attribute `MySDK.v3_2.MySession.root` will be populated with information such as your name,
your phone number, your avatar, your enterprise name and ID etc. This `user` is the root object of everything as all subsequent
calls need to be done in the context of your account (for instance, your `/enterprises` are different from another account's `/enterprises`)

It is also possible to create sub sessions with the python statement `with`:

    #!python
    cspsession = MySession(username="user", password="secret", enterprise="organization", api_url="https://server")
    adminsession = MySession(username="admin", password="secret", enterprise="enterprise", api_url="https://server")

    cspsession.start()

    # this part of the code will use the CSP root user

    with adminsession.start():
        # this code block will be executed as admin of `enterprise`

    # back to csp root session

> You **must** use `start()` when using the `with` statement, even if the session has already been started in the main context.



## NURESTObject

`bambou.NURESTObject` is the parent class of all `MySDK` entities.


### ReST Names

All `bambou.NURESTObject` subclasses implements a given method that will return the actual ReST name of the objects. For instance, the ReST name of an Unicorn object is `unicorn`.

These names are used to forge the correct URI when doing CRUD operations on them.

> ReST names can be used as unique resource identifier for a given object.

> ReST names are auto generated. You never need to manually define them.


### ReST API URI Generation

`bambou.NURESTObject` is able to forge all the URI needed to interact with the server through the ReST API.

For instance, if an object with a ReST name set to `object` needs to get the list of children with ReST name set to `subobject`, `Bambou` will use the following endpoint URL:

    `GET {api_base_url}/objects/{id}/subobjects`

If an object with a ReST name set to `entity` needs to fetch itself, the generated URL will be

    `GET {api_base_url}/entities/{id}`


> `Bambou` automagically deals with plurals.

> The ReST base URL is pulled from the current active `bambou.NURESTSession`.

> URI are auto generated. You never need to deal with them manually.


### Exposing ReST Attributes

Exposed attributes will be converted and sent to the server when you do CRUD operations. That way, if an object has an attribute `name`, it can be marked as a ReST attribute.

When saving the object, the value of `name` will be put into the generated JSON structure that will be sent to the server, or automatically populated from a JSON structure that is coming from the server.

Not only the attribute can be exposed, but also its type and other informations like if it is read only, its allowed values, its format, its default value and so on.

> exposing ReST Attributes is auto generated. You never need to manually expose new attributes.


### CRUD Operations

`bambou.NURESTObject` allows to perform all sorts of CRUD operations.

* `bambou.NURESTObject.fetch`
* `bambou.NURESTObject.save`
* `bambou.NURESTObject.delete`
* `bambou.NURESTObject.create_child`
* `bambou.NURESTObject.assign`
* `bambou.NURESTObject.instantiate_child`

> All these methods require the current `bambou.NURESTObject` to have a valid `bambou.NURESTObject.ID`.

> You may notice that there is no creation method. Creation is always happening from a parent object and is done using `create_child`.

> You may notice that an optional parameter `callback` is present. This is because `MySDK` can work completely asynchronously.


### Converting to and from a Python Dictionary

`bambou.NURESTObject` allows quick and easy conversion from and to python dictionaries

* `bambou.NURESTObject.from_dict`
* `bambou.NURESTObject.to_dict`

> you never need to process to the actual JSON conversion when sending info to the server. `bambou.NURESTConnection` will do that automatically, but you can use these methods to print an object, or copy information of an object into one another.



## NURESTFetcher

`bambou.NURESTFetcher` is a class allowing a `bambou.NURESTObject` to fetch its children. All `bambou.NURESTObject` have one or more fetchers, unless it's a final object in the  model hierarchy. `bambou.NURESTFetcher` provides a lot of possibility regarding the way you want to get a given children list. It can deal with simple object fetching, pagination, filtering, request headers, grouping etc.


### Fetching Children List

`bambou.NURESTFetcher` has three importants methods:

* `bambou.NURESTFetcher.fetch`
* `bambou.NURESTFetcher.get`
* `bambou.NURESTFetcher.get_first`


### Discussion about Fetchers

Fetcher is a powerfull concept that makes the process of getting child objects completely generic and code friendly. `bambou.NURESTObject` provides methods that allow to deal programatically with its fetchers in a completely generic way.

* `bambou.NURESTObject.fetcher_for_rest_name`
* `bambou.NURESTObject.fetchers`
* `bambou.NURESTObject.children_rest_names`

This allows complete abstract programatic operations on any objects.

For instance, the following function will create a new `MySDK.v3_2.Metadata` to the entire hierarchy of children from a given object that has been created after a certain date:

    #!python
    def apply_metatada_to_all_children(root_object, metadata, filter=None):

        # Loop on all declared children fetchers
        for fetcher in root_object.fetchers:

            # Fetch the list of the children
            children = fetcher.get(filter=filter)

            # Loop on all fetched children
            for child in children:

                # Add the metadata to the current children
                child.create_child(metadata)

                # Start over recursively on the children of the current child
                apply_metadata_to_all_children(child, metadata)


    enterprise = Enterprise(id="xxxx-xxxx-xxx-xxxx")
    metadata = Metadata(name="my metadata", blob="hello world!")

    apply_metadata_to_all_children(enterprise, metadata, filter="creationDate > '01-01-2015'")



## NURESTPushCenter

The API supports client side push through a long polling connection. ReST clients can connect to that channel and will get a notification as soon as he or someone else in the system changes something. This events are filtered by permissions, which means that if someone change a property of an object you cannot see, you won't get notified. `MySDK` provides the `bambou.NURESTPushCenter`, which encapsulates all the logic to deal with the event channel. It runs in its own thread and will call registered callbacks when it receives a push.

A `bambou.NURESTPushCenter` is automatically created with each `bambou.NURESTSession` and it is available from the attribute `bambou.NURESTSession.push_center`.

    #!python
    session = MySession(username="user", password="secret", enterprise="organization", api_url="https://server")
    session.start()
    session.push_center.start()

> You need to explicitely start the push center.


### Using the NURESTPushCenter

Only the following methods are important:

* `bambou.NURESTPushCenter.start`
* `bambou.NURESTPushCenter.add_delegate`
* `bambou.NURESTPushCenter.remove_delegate`


### Example

Here is a really simple code sample that will print the push data on every push:

    #!python
    from MySDK import *
    from pprint import pprint
    from time import sleep

    session = MySession(username="csproot", password="secret", enterprise="csp", api_url="https://server")
    session.start()

    def on_receive_push(data):
        pprint(data);

    session.push_center.add_delegate(on_receive_push);
    session.push_center.start()

    # default stupid run loop. don't do that in real life :)
    while True:
        sleep(1000)


## Conclusion

Now you know the basics of `Bambou` and so, of the `MySDK`. Remember that all objects in `MySDK` are subclasses of `bambou.NURESTObject` so they **all** work exactly the same.

There is a lot more to know about `Bambou` like the asynchronous mode, auto model parsing, easy controllers creation thanks introspection and so on. We'll cover this in a different advanced section.
"""

try:
    import requests
    requests.packages.urllib3.disable_warnings()
except:
    pass

import logging

bambou_logger = logging.getLogger('bambou')
pushcenter_logger = logging.getLogger('pushcenter')

try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

bambou_logger.addHandler(NullHandler())

__all__ = ['NURESTRootObject', 'NURESTConnection', 'NURESTModelController', 'NURESTFetcher', 'NURESTLoginController', 'NURESTObject', 'NURESTPushCenter', 'NURESTRequest', 'NURESTResponse', 'NURESTSession', 'BambouConfig']

from bambou.nurest_session import NURESTSession
from bambou.nurest_root_object import NURESTRootObject
from bambou.nurest_connection import NURESTConnection
from bambou.nurest_fetcher import NURESTFetcher
from bambou.nurest_login_controller import NURESTLoginController
from bambou.nurest_object import NURESTObject
from bambou.nurest_push_center import NURESTPushCenter
from bambou.nurest_request import NURESTRequest
from bambou.nurest_response import NURESTResponse
from bambou.nurest_modelcontroller import NURESTModelController
from bambou.config import BambouConfig




# -*- coding: utf-8 -*-

# Copyright 2014 Alcatel-Lucent USA Inc.

from builtins import str
from builtins import object
from .nurest_login_controller import NURESTLoginController
from .nurest_push_center import NURESTPushCenter
from bambou import bambou_logger
from contextlib import contextmanager
import requests
import sys


class NURESTSession(object):

    """ Authenticated sessions used for sending ReST calls

        The session holds the credential information for a particular
        user. It is used by :class:`bambou.NURESTConnection` to get
        the authentication information needed to send the request.

        NURESTSession supports the `with` statement.

        Note:
            NURESTSession *must* be subclassed, and the subclass *must* implement :class:`bambou.NURESTSession.create_root_object`

        Example:
            >>> mainsession =  NUMySession(username="csproot", password="csproot", enterprise="csp", api_url="https://vsd:8443")
            >>> othersession = NUMySession(username="user", password="password", enterprise="ent", api_url="https://vsd:8443")
            >>>
            >>> mainsession.start()
            >>> mainsession.user.entities.get()
            [<NUEntity at 1>, <NUEntity at 2>, <NUEntity at 3>]
            >>>
            >>> with othersession.start() as session:
            >>>     session.user.entities.get()
            [<NUEntity at 2>]
    """
    
    current_context = None
    current_session = None
    session_stack = []

    def __init__(self, username, password, enterprise, api_url, api_prefix, version, certificate=None):
        """ Initializes a new sesssion

            Args:
                username (string): the username
                password (string): the password
                enterprise (string): the enterprise
                api_url (string): the url to the api
                version (string): the version of the api to target

            Example:
                >>> mainsession =  NUMySession(username="csproot", password="csproot", enterprise="csp", api_url="https://vsd:8443")

        """
        self._root_object = None
        self._login_controller = NURESTLoginController()
        self._login_controller.user = username
        self._login_controller.password = password
        self._login_controller.certificate = certificate
        self._login_controller.user_name = username
        self._login_controller.enterprise = enterprise
        self._login_controller.url = '%s/%s/v%s' % (api_url, api_prefix, str(version).replace('.', '_'))

        self._push_center = NURESTPushCenter()
        self._push_center.url = self._login_controller.url
        self.requests_session = requests.Session()

    # Class Methods

    @classmethod
    def get_current_session(cls):
        """
            Get the current session

            Returns:
                (bambou.NURESTSession): the current session

        """
        if len(NURESTSession.session_stack) > 0:
            return NURESTSession.session_stack[-1]
        else:
            return NURESTSession.current_session

    # Properties

    @property
    def push_center(self):
        """
            Returns the :class:`bambou.NURESTPushCenter` of the current session

            Note:
                Use this method to start and stop receiving push notifications
        """
        return self._push_center

    @property
    def login_controller(self):
        """
            Returns the :class:`bambou.NURESTLoginController` of the current session

            Note:
                You should not need to use this method. It's used automatically when needed
        """
        return self._login_controller

    @property
    def root_object(self):
        """
            Returns the root object of the session

            Returns:
                (bambou.NURESTRootObject): the root object
        """
        return self._root_object

    @property
    def is_impersonating(self):
        """ Returns True if the session is currently impersonating
            a root object

            Returns:
                (bool): a boolean that indicate if the session is impersonating a root object

        """
        return self._login_controller.is_impersonating

    # Methods

    def create_root_object(self):
        """
            Create a :class:`bambou.NURESTRootObject`.

            This method *MUST* be overriden by subclasses in order to provide
            a valid :class:`bambou.NURESTRootObject`.

            Returns:
                A instance of a subclass of :class:`bambou.NURESTRootObject`
        """
        raise NotImplementedError('%s must define method def create_root_object(self).' % self)

    def _authenticate(self):

        if self._root_object is None:
            self._root_object = self.create_root_object()
            self._root_object.fetch()

        self.login_controller.api_key = self._root_object.api_key
        bambou_logger.debug("[NURESTSession] Started session with username %s in enterprise %s" % (self.login_controller.user, self.login_controller.enterprise))

    def start(self):
        """
            Starts the session.

            Starting the session will actually get the API key of the current user
        """

        if NURESTSession.session_stack:
            bambou_logger.critical("Starting a session inside a with statement is not supported.")
            raise Exception("Starting a session inside a with statement is not supported.")

        NURESTSession.current_session = self

        self._authenticate()
        return self

    def reset(self):
        """
            Resets the session.

            Resetting the session will flush the API stored API key. Any additional calls will require to call start, and a
            /me request will be reissued.
        """

        self._root_object = None
        self.login_controller.api_key = None

    def impersonate(self, username, enterprise):
        """
            Change the session to impersonate a user within an enterprise

            Args:
                username (string): name of the user to impersonate
                enterprise (string): name of the enterprise
        """
        self._login_controller.impersonate(username, enterprise)

    def stop_impersonate(self):
        """
            Stop impersonating a user

        """
        self._login_controller.stop_impersonate()

    def equals(self, session):
        """ Verify if the current session equals the given parameter

            Notes:
                Verification is based on username, enterprise, api_url and its version.

            Args:
                session(bambou.NURESTSession): the session to compare with

            Returns:
                (bool): True if session are equal
        """
        return self.login_controller.equals(session.login_controller)

    def is_current_session(self):
        """ Verify if the session is the current.

            Returns:
                (bool): True if the session is the current
        """
        current_session = NURESTSession.get_current_session()
        return current_session and self.equals(current_session)

    def __enter__(self):
        NURESTSession.session_stack.append(self)
        return self

    def __exit__ (self, exc_type, exc_value, traceback):
        NURESTSession.session_stack.pop(-1)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015, Alcatel-Lucent Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its contributors
#       may be used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from builtins import object
class NURESTModelController(object):
    """ Access any object via its remote name """

    _model_rest_name_registry = {}
    _model_resource_name_registry = {}

    @classmethod
    def register_model(cls, model):
        """
            Register a model class according to its remote name

            Args:
                model: the model to register
        """

        rest_name = model.rest_name
        resource_name = model.resource_name

        if rest_name not in cls._model_rest_name_registry:
            cls._model_rest_name_registry[rest_name] = [model]
            cls._model_resource_name_registry[resource_name] = [model]

        elif model not in cls._model_rest_name_registry[rest_name]:
            cls._model_rest_name_registry[rest_name].append(model)
            cls._model_resource_name_registry[resource_name].append(model)

    @classmethod
    def get_all_models(cls):
        """ Retrieve all models

            Returns:
                A list of all models.
                An empty list if no entries found for the remote name
        """

        return list(cls._model_rest_name_registry.values())

    # Get models by rest_name

    @classmethod
    def get_models_with_rest_name(cls, rest_name):
        """ Retrieve all models from a given rest name

            Args:
                rest_name: the rest name entry

            Returns:
                A list of models corresponding to remote name arg.
                An empty list if no entries found for the remote name
        """

        if rest_name in cls._model_rest_name_registry:
            return cls._model_rest_name_registry[rest_name]

        return []

    @classmethod
    def get_first_model_with_rest_name(cls, rest_name):
        """ Get the first model corresponding to a rest_name

            Args:
                rest_name: the rest name
        """

        models = cls.get_models_with_rest_name(rest_name)

        if len(models) > 0:
            return models[0]

        return None

    # Get models by resource_name

    @classmethod
    def get_models_with_resource_name(cls, resource_name):
        """ Retrieve all models from a given resource_name

            Args:
                resource_name: the resource name of the model

            Returns:
                A list of models corresponding to remote name arg.
                An empty list if no entries found for the remote name
        """

        if resource_name in cls._model_resource_name_registry:
            return cls._model_resource_name_registry[resource_name]

        return []

    @classmethod
    def get_first_model_with_resource_name(cls, resource_name):
        """ Get the first model corresponding to a resource_name

            Args:
                resource_name: the resource name
        """

        models = cls.get_models_with_resource_name(resource_name)

        if len(models) > 0:
            return models[0]

        return None

    # will be deprecated
    @classmethod
    def get_models(cls, rest_name):
        """
        """
        return cls.get_models_with_rest_name(rest_name)

    # will be deprecated
    @classmethod
    def get_first_model(cls, rest_name):
        """
        """
        return cls.get_first_model_with_rest_name(rest_name)
